import subprocess

def install_library(library):
	try:
		result = subprocess.run(['pip', 'install', library], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		return result.returncode == 0
	except subprocess.CalledProcessError:
		return False
	

def uninstall_library(library):
    try:
        result = subprocess.run(['pip', 'uninstall', '-y', library], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return result.returncode == 0
    except subprocess.CalledProcessError:
        return False


import requests
from bs4 import BeautifulSoup

try:
    from ..character import set_website_content
except:
    from gpt_computer_assistant.character import set_website_content


def train(url: str) -> bool:
    try:
        # Go to url and extract these elements
        meta_properties = [
            "og:description",
            "og:site_name",
            "og:title",
            "og:type",
            "og:url",
        ]

        # Fetch the webpage content
        response = requests.get(url)
        soup = BeautifulSoup(response.content, "html.parser")

        # Extract the meta tags
        meta_tags = soup.find_all("meta")

        # Initialize the data dictionary
        data = {}

        # Loop through the meta tags and extract the content
        for tag in meta_tags:
            if tag.get("property") in meta_properties:
                data[tag.get("property")] = tag.get("content")



        # Also add the other useful information texts from the webpage
        data["title"] = soup.title.string
        data["h1"] = soup.h1.string
        data["p"] = soup.p.string

        
        text = soup.get_text(separator='\n', strip=True)
        
        data["text"] = text

        data["url"] = url


        # Now create an string with good looking like this
        # Title: {title}


        the_string = ""

        for key, value in data.items():
            the_string += f"{key}: {value}\n"



        set_website_content(the_string)

        return True

    except Exception as e:
        return e





from openai import OpenAI
from langchain_openai import ChatOpenAI
from langchain_community.chat_models import ChatOllama
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_groq import ChatGroq
import copy

try:
    from .utils.db import load_api_key, load_openai_url, load_model_settings, load_groq_api_key, load_google_api_key
    from .custom_callback import customcallback
    from .llm_settings import llm_settings
except ImportError:
    from utils.db import load_api_key, load_openai_url, load_model_settings, load_groq_api_key, load_google_api_key
    from custom_callback import customcallback
    from llm_settings import llm_settings



the_callback = customcallback(strip_tokens=False, answer_prefix_tokens=["Answer"])



def get_model(high_context=False):
    the_model = load_model_settings()
    the_api_key = load_api_key()
    the_groq_api_key = load_groq_api_key()
    the_google_api_key = load_google_api_key()
    the_openai_url = load_openai_url()

    def open_ai_base(high_context):
        if the_openai_url == "default":
            true_model = the_model
            if high_context:
                true_model = "gpt-4-turbo"
            return {"model": true_model, "api_key": the_api_key, "max_retries":15, "streaming":True, "callbacks":[the_callback]}
        else:
            return {"model": the_model, "api_key": the_api_key, "max_retries":15, "streaming":True, "callbacks":[the_callback], "base_url": the_openai_url}

    args_mapping = {
        ChatOpenAI: open_ai_base(high_context=high_context),
        ChatOllama: {"model": the_model},
        ChatGroq: {"temperature": 0, "model_name": the_model.replace("-groq", ""), "groq_api_key": the_openai_url},
        ChatGoogleGenerativeAI:{"model": the_model, "google_api_key": the_google_api_key}
    }


    model_mapping = {}

    for model_name, model_args in llm_settings.items():
        the_tuple = None
        if model_args["provider"] == "openai":
            the_tuple = (ChatOpenAI, args_mapping[ChatOpenAI])
        elif model_args["provider"] == "ollama":
            the_tuple = (ChatOpenAI, {"api_key": "ollama", "base_url":"http://localhost:11434/v1", "model": model_name})
        elif model_args["provider"] == "google":
            the_tuple = (ChatGoogleGenerativeAI, args_mapping[ChatGoogleGenerativeAI])
        elif model_args["provider"] == "groq":
            the_tuple = (ChatGroq, args_mapping[ChatGroq])
        
        if the_tuple:
            model_mapping[model_name] = the_tuple

        


    model_class, args = model_mapping[the_model]
    return model_class(**args) if model_class else None


def get_client():
    the_api_key = load_api_key()
    the_openai_url = load_openai_url()
    if the_openai_url == "default":
        return OpenAI(api_key=the_api_key)
    else:
        return OpenAI(api_key=the_api_key, base_url=the_openai_url)



import textwrap
import requests

import time
import sys
from upsonic import Tiger


the_upsonic_ = None


def the_upsonic():
    global the_upsonic_

    if not the_upsonic_:
        the_upsonic_ = Tiger()
    
    return the_upsonic_

class Remote_Client:
    def __init__(self, url):
        self.url = url

        if self.status != True:
            raise Exception("The server is not running")


    def send_request(self, path, data, files=None, dont_error=False):
        try:
            if files == None:
                response = requests.post(self.url+path, json=data)
            else:
                response = requests.post(self.url+path, data=data, files=files)
            if response.status_code != 200:
                try:
                    print(response.json())
                except:
                    print(response.text)
                
                raise Exception("Request failed", response.status_code, path)
            return response.json()
        except Exception as e:
            if dont_error:
                return {"response": str(e)}
            else:
                raise e


    @property
    def status(self):
        data = {}
        response = self.send_request("/status", data, dont_error=True)
        return response["response"]

    def input(self, text:str, screen:bool=False, talk:bool=False) -> str:
        data = {"text": text, "screen": str(screen).lower(), "talk": str(talk).lower()}
        response = self.send_request("/input", data)
        return response["response"]

    def just_screenshot(self) -> str:
        data = {}
        response = self.send_request("/screenshot", data)
        return response["response"]

    def talk(self, text:str) -> str:
        data = {"text": text}
        response = self.send_request("/tts", data)
        return response["response"]

    def profile(self, profile:str) -> str:
        data = {"profile": profile}
        response = self.send_request("/profile", data)
        return response["response"]

    def reset_memory(self) -> str:
        response = self.send_request("/reset_memory", {})
        return response["response"]

    def enable_predefined_agents(self) -> str:
        response = self.send_request("/activate_predefined_agents", {})
        return response["response"]

    def disable_predefined_agents(self) -> str:
        response = self.send_request("/deactivate_predefined_agents", {})
        return response["response"]

    def enable_online_tools(self) -> str:
        response = self.send_request("/activate_online_tools", {})
        return response["response"]

    def disable_online_tools(self) -> str:
        response = self.send_request("/deactivate_online_tools", {})
        return response["response"]


    def change_name(self, new_name:str) -> str:
        data = {"new_name": new_name}
        response = self.send_request("/change_name", data)
        return response["response"]
    
    def change_developer(self, new_developer:str) -> str:
        data = {"new_developer": new_developer}
        response = self.send_request("/change_developer", data)
        return response["response"]


    def install_library(self, library:str) -> str:
        data = {"library": library}
        response = self.send_request("/library_install", data)
        return response["response"]
    
    def uninstall_library(self, library:str) -> str:
        data = {"library": library}
        response = self.send_request("/library_uninstall", data)
        return response["response"]


    def custom_tool(self, func):
        the_code = textwrap.dedent(the_upsonic().extract_source(func))
        # Remove the first line

        if the_code.startswith("@remote.custom_tool"):
            the_code = the_code[the_code.find("\n")+1:]
        
        data = {"code": the_code}
        response = self.send_request("/custom_tool", data)
        return response["response"]


    def top_bar_animation(self, text):
        data = {"text": text}
        response = self.send_request("/top_bar_activate", data)


    def stop_top_bar_animation(self, text):
        data = {"text": text}
        response = self.send_request("/top_bar_deactivate", data)


    def boop(self):
        data = {}
        response = self.send_request("/boop_sound", data)


    def ask(self, question, wait_for_answer=None):
        data = {"question":question, "wait_for_answer":wait_for_answer}
        response = self.send_request("/ask_to_user", data)
        return response["response"]


    def set_text(self, text):
        data = {"text": text}
        response = self.send_request("/set_text", data)
        return response["response"]

    class OperationContext:
        def __init__(self, client, text):
            self.client = client
            self.text = text

        def __enter__(self):
            self.client.top_bar_animation(self.text)
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.client.stop_top_bar_animation(self.text)

    def operation(self, text):
        return self.OperationContext(self, text)


    def set_background_color(self, r,g,b):
        data = {"color": f"{r}, {g}, {b}"}
        response = self.send_request("/set_background_color", data)
        return response["response"]
    
    def set_opacity(self, opacity):
        data = {"opacity": opacity}
        response = self.send_request("/set_opacity", data)
        return response["response"]

    def set_border_radius(self, radius):
        data = {"radius": radius}
        response = self.send_request("/set_border_radius", data)
        return response["response"]


    def collapse(self):
        data = {}
        response = self.send_request("/collapse", data)
        return response["response"]
    

    def expand(self):
        data = {}
        response = self.send_request("/expand", data)
        return response["response"]
    

    def save_openai_api_key(self, openai_api_key):
        data = {"openai_api_key": openai_api_key}
        response = self.send_request("/save_openai_api_key", data)
        return response["response"]

    def save_openai_url(self, openai_url):
        data = {"openai_url": openai_url}
        response = self.send_request("/save_openai_url", data)
        return response["response"]
    
    def save_model_settings(self, model_name):
        data = {"model_name": model_name}
        response = self.send_request("/save_model_settings", data)
        return response["response"]

    def save_groq_api_key(self, groq_api_key):
        data = {"groq_api_key": groq_api_key}
        response = self.send_request("/save_groq_api_key", data)
        return response["response"]

    def save_google_api_key(self, google_api_key):
        data = {"google_api_key": google_api_key}
        response = self.send_request("/save_google_api_key", data)
        return response["response"]
    
    def save_tts_model_settings(self, model_name):
        data = {"model_name": model_name}
        response = self.send_request("/save_tts_model_settings", data)
        return response["response"]

    def save_stt_model_settings(self, model_name):
        data = {"model_name": model_name}
        response = self.send_request("/save_stt_model_settings", data)
        return response["response"]


    

    def show_logo(self):
        data = {}
        response = self.send_request("/show_logo", data)
        return response["response"]
    
    def hide_logo(self):
        data = {}
        response = self.send_request("/hide_logo", data)
        return response["response"]


    def custom_logo(self, logo_path):
        data = {}
        files = {"logo": open(logo_path, "rb")}
        response = self.send_request("/custom_logo_upload", data, files)
        return response["response"]
    
    def default_logo(self):
        data = {}
        response = self.send_request("/default_logo", data)
        return response["response"]



    def activate_long_gca(self):
        data = {}
        response = self.send_request("/activate_long_gca", data)
        return response["response"]
    
    def deactivate_long_gca(self):
        data = {}
        response = self.send_request("/deactivate_long_gca", data)
        return response["response"]



    def train(self, url):
        data = {"url": url}
        response = self.send_request("/train", data)
        return response["response"]



    def wait(self, second):
        time.sleep(second)




remote = Remote_Client("http://localhost:7541")



from .utils.db import agents

class Agent:
    """
    Represents an agent within the system.

    This class defines an agent with a specific role, goal, and backstory. Upon initialization,
    the agent is added to the global list of agents.

    Attributes:
    - role (str): The role of the agent.
    - goal (str): The goal or objective of the agent.
    - backstory (str): The backstory or history of the agent.

    Methods:
    - __init__(role, goal, backstory): Initializes the Agent object and adds it to the global list of agents.

    Global Variables:
    - agents (list): A global list containing information about all agents in the system.
    """
    def __init__(self, role, goal, backstory):
        """
        Initializes a new Agent object and adds it to the global list of agents.

        Parameters:
        - role (str): The role of the agent.
        - goal (str): The goal or objective of the agent.
        - backstory (str): The backstory or history of the agent.

        Returns:
        None
        """
        global agents
        agents.append({"role": role, "goal": goal, "backstory": backstory})



from langchain_core.messages import SystemMessage
try:
    from .chat_history import *
    from ..llm_settings import first_message
except ImportError:
    from agent.chat_history import *
    from llm_settings import first_message

def llm_history_oiginal():

    return [
    SystemMessage(
        content=[
            {
                "type": "text",
                "text": first_message(),
            }
        ]
    ),    
]



def string_to_function(func_str):
	# Define a local dictionary to execute the function string
	local_dict = {}
	
	# Execute the function string in the local dictionary
	exec(func_str, globals(), local_dict)
	
	# Extract the function name
	func_name = func_str.split('(')[0].split()[-1]
	
	# Return the function object from the local dictionary
	return local_dict[func_name]




try:
    from ..llm import *
    from ..utils.db import *
    from .tts_providers.openai import tts_openai
    from .tts_providers.microsoft_local import tts_microsoft_local
except ImportError:
    from llm import *
    from utils.db import *
    from audio.tts_providers.openai import tts_openai
    from audio.tts_providers.microsoft_local import tts_microsoft_local

import os
import hashlib
import random
import threading


def is_local_tts_available():
    try:
        from transformers import pipeline
        from datasets import load_dataset
        import soundfile as sf

        import torch
        return True
    except:
        return False


def is_openai_tts_available():

    the_model = load_model_settings()
    if llm_settings[the_model]["provider"] == "openai":
        if load_api_key() != "CHANGE_ME":
            return True
    return False



supported_openai_speakers = ["fable"]

def random_model(exclude):
    models = supported_openai_speakers.copy()
    models.remove(exclude)
    return random.choice(models)

def generate_speech_chunk(text_chunk, index, voice, results):
    sha = hashlib.sha256(text_chunk.encode()).hexdigest()
    location = os.path.join(artifacts_dir, f"{sha}.mp3")

    if os.path.exists(location):
        results[index] = location
    else:
        the_model = load_model_settings()
        tts_setting = load_tts_model_settings()
        if tts_setting == "openai" : 
            tts_openai(voice, text_chunk, location)
        
        if tts_setting == "microsoft_local":
            if not is_local_tts_available():
                print("Please install gpt-computer-assistant[local_tts] to use local TTS")
            else:
                tts_microsoft_local(text_chunk, location)
                
            
                
        results[index] = location

def split_text_to_sentences(text, max_chunk_size=300):
    """Splits text into sentences and ensures chunks do not exceed max_chunk_size."""
    sentences = text.split('.')
    chunks = []
    current_chunk = ""

    for sentence in sentences:
        sentence = sentence.strip()
        if len(current_chunk) + len(sentence) + 1 <= max_chunk_size:
            current_chunk += (sentence + '. ')
        else:
            chunks.append(current_chunk.strip())
            current_chunk = sentence + '. '

    if current_chunk:
        chunks.append(current_chunk.strip())

    return chunks

def text_to_speech(text):
    text_chunks = split_text_to_sentences(text)

    threads = []
    results = [None] * len(text_chunks)

    initial_voice = random.choice(supported_openai_speakers)

    for i, chunk in enumerate(text_chunks):
        voice = initial_voice if i % 2 == 0 else random_model(initial_voice)  # Alternate voices
        thread = threading.Thread(
            target=generate_speech_chunk,
            args=(chunk, i, voice, results)
        )
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join()

    mp3_files = [result for result in results if result is not None]

    return mp3_files




try:
    from ..utils.db import *
    from ..tooler import *
    from ..display_tools import *
    from ..teams import *
    from ..llm_settings import each_message_extension, llm_settings
except ImportError:

    from utils.db import *

    from tooler import *
    from display_tools import *
    from teams import *
    from llm_settings import each_message_extension, llm_settings



custom_tools = []



def load_tiger_tools():
    try:
        from upsonic import Tiger
        tools = Tiger()
        tools.enable_auto_requirements = True
        tools = tools.langchain()
        return tools
    except:
        return False


def load_default_tools():
    from ..standard_tools import get_standard_tools
    return get_standard_tools()




cached_tiger_tools = None

def get_tiger_tools():
    global cached_tiger_tools
    if cached_tiger_tools is None:
        cached_tiger_tools = load_tiger_tools()
    return cached_tiger_tools

if is_online_tools_setting_active():
    get_tiger_tools()

def get_tools():

    model = load_model_settings()




    if not llm_settings[model]["tools"]:
        from ..standard_tools import get_current_time
        from ..tooler import tool
        return []


    if is_online_tools_setting_active():
        tools = get_tiger_tools()
        if not tools:
            tools = load_default_tools()
    else:
        tools = load_default_tools()

    return tools


from bs4 import BeautifulSoup
import requests
import re
from urllib.parse import urljoin
import datetime
import traceback

try:
    from .tooler import tool
    from .top_bar_wrapper import wrapper
except:
    from tooler import tool
    from top_bar_wrapper import wrapper

_standard_tools_ = {}

def register_tool(func):
    if func.__name__ not in _standard_tools_:
        _standard_tools_[func.__name__] = tool(func)
    return func
@register_tool
@wrapper
def read_website(url: str, max_content_length: int = 5000) -> dict:
    """
    Read the content of a website and return the title, meta data, content, and sub-links.
    """
    try:
        response = requests.get(url)
        response.raise_for_status()
        html = response.text
    except requests.RequestException as e:
        return {"error": f"Failed to retrieve the website content: {e}"}

    soup = BeautifulSoup(html, "html.parser")

    meta_properties = [
        "og:description",
        "og:site_name",
        "og:title",
        "og:type",
        "og:url",
        "description",
        "keywords",
        "author"
    ]
    meta = {}
    for property_name in meta_properties:
        tag = soup.find("meta", property=property_name) or soup.find("meta", attrs={"name": property_name})
        if tag:
            meta[property_name] = tag.get("content", "")

    for ignore_tag in soup(["script", "style"]):
        ignore_tag.decompose()

    title = soup.title.string.strip() if soup.title else ""
    content = soup.body.get_text(separator="\n") if soup.body else ""

    links = []
    for a in soup.find_all("a", href=True):
        link_url = urljoin(url, a["href"])
        links.append({"title": a.text.strip(), "link": link_url})

    content = re.sub(r"[\n\r\t]+", "\n", content)
    content = re.sub(r" +", " ", content)
    content = re.sub(r"[\n ]{3,}", "\n\n", content)
    content = content.strip()

    if len(content) > max_content_length:
        content = content[:max_content_length].rsplit(' ', 1)[0] + '...'

    return {"meta": meta, "title": title, "content": content, "sub_links": links}


@register_tool
@wrapper
def google(query: str, max_number: int = 20) -> list:
    """
    Search the query on Google and return the results.
    """
    try:
        from googlesearch import search as gsearch
        return list(gsearch(query, stop=max_number))
    except:
        return "An exception occurred"    


@register_tool
@wrapper
def duckduckgo(query: str, max_number: int = 20) -> list:
    """
    Search the query on DuckDuckGo and return the results.
    """
    try:
        from duckduckgo_search import DDGS
        return [result["href"] for result in DDGS().text(query, max_results=max_number)]
    except:
        return "An exception occurred"



@register_tool
@wrapper
def copy(text: str):
    """
    Copy the text to the clipboard.
    """
    import pyperclip
    pyperclip.copy(text)
    pyperclip.copy(text)


@register_tool
@wrapper
def open_url(url) -> bool:
    """
    Open the URL in the default web browser.

    :param url: str:
    """
    import webbrowser

    try:
        webbrowser.open(url)
        return True
    except:
        return False
        return False

@register_tool
@wrapper
def sleep(seconds: int):
    """
    Sleep for the given number of seconds.
    """
    import time
    time.sleep(seconds)



@register_tool
@wrapper
def keyboard_write(text: str):
    """
    Write the text using the keyboard.
    """
    import pyautogui
    pyautogui.write(text)

@register_tool
@wrapper
def keyboard_press(key: str):
    """
    Press the key using the keyboard.
    """
    import pyautogui
    pyautogui.press(key)
    pyautogui.press(key)



from langchain_experimental.utilities import PythonREPL

the_py_client = PythonREPL()

@register_tool
@wrapper
def python_repl(code: str) -> str:
    """
    Run and return the given python code in python repl
    """
    return the_py_client.run(code)

@register_tool
@wrapper
def app_open(app_name: str) -> bool:
    """
    Opens the native apps.
    """
    try:
        from AppOpener import open
        open(app_name, throw_error=True)
        return True
    except:
        try:
            from MacAppOpener import open
            open(app_name)
        except:
            return False

@register_tool
@wrapper
def app_close(app_name: str) -> bool:
    """
    Closes the native apps.
    """
    try:
        from AppOpener import close
        close(app_name, throw_error=True)
        return True
    except:
        try:
            from MacAppOpener import open
            close(app_name)
        except:
            return False



@register_tool
@wrapper
def get_current_time() -> str:
    """
    Get the current time in ISO format.
    """
    return datetime.datetime.now().isoformat()



@register_tool
@wrapper
def turn_off_wifi() -> bool:
    """
    Turn off the wifi.
    """
    try:
        from pywifi import ControlPeripheral
        wifi = ControlPeripheral()
        wifi.disable()
        return True
    except:

        return False
    
@register_tool
@wrapper
def turn_on_wifi() -> bool:
    """
    Turn on the wifi.
    """
    try:
        from pywifi import ControlPeripheral
        wifi = ControlPeripheral()
        wifi.enable()
        return True
    except:
        
        return False
    

@register_tool
@wrapper
def connect_wifi(ssid: str, password: str) -> bool:
    """
    Connect to the wifi with the given ssid and password.
    """
    try:
        from pywifi import ControlConnection

        # Arguments passed during object instantiation
        controller = ControlConnection(wifi_ssid=ssid, wifi_password=password)
        controller.wifi_connector()
        return True
    except:
        return False




@register_tool
@wrapper
def ask_to_user(question:str, wait_for_answer:str=None) -> str:
    """
    Its ask to the user for your question and return the answer
    """
    try:

        try:
            from .agent.process import tts_if_you_can
            from .audio.record import quick_speech_to_text
        except:
            from agent.process import tts_if_you_can
            from audio.record import quick_speech_to_text

        print("TTS")
        tts_if_you_can(question, bypass_other_settings=True, not_threaded=True)
        print("TTS END")

        if wait_for_answer:
            return quick_speech_to_text(wait_for_answer)
        else:
            return quick_speech_to_text()
    except:
        traceback.print_exc()
        return False


def get_standard_tools():
    print("Tool len", len(_standard_tools_))
    last_list = [_standard_tools_[each] for each in _standard_tools_]
    return last_list



if __name__ == "__main__":
    print(ask_to_user("What is your age"))


# Create a python api and start_api function via flask

from flask import Flask, request, jsonify
import threading
import time

from werkzeug.serving import make_server

app = Flask(__name__)


@app.route("/status", methods=["POST"])
def status():
    return jsonify({"response": True})



@app.route("/input", methods=["POST"])
def input():
    """
    This function receives input from the user and returns the response.
    """
    data = request.json
    text = data["text"]
    screen = data["screen"]
    talk = data["talk"]
    print("Input:", text)
    from .gpt_computer_assistant import the_main_window, the_input_box

    firsst_text = the_input_box.toPlainText()

    original_tts = the_main_window.tts_available

    if talk == "true":
        the_main_window.tts_available = True
        the_main_window.manuel_stop = True

    if screen != "true":
        the_main_window.button_handler.input_text(text)
    else:
        the_main_window.button_handler.input_text_screenshot(text)


    while the_input_box.toPlainText() == firsst_text:
        time.sleep(0.3)

    while the_input_box.toPlainText().startswith("System:"):
        time.sleep(0.3)

    while not the_main_window.state == "idle":
        time.sleep(0.3)

    response = the_input_box.toPlainText()



    the_main_window.tts_available = original_tts

    return jsonify({"response": response})


@app.route("/screenshot", methods=["POST"])
def screenshot():
    """
    This function receives a screenshot from the user and returns the response.
    """
    from .gpt_computer_assistant import the_main_window, the_input_box
    firsst_text = the_input_box.toPlainText()
    the_main_window.button_handler.just_screenshot()

    while the_input_box.toPlainText() == firsst_text:
        time.sleep(0.3)

    while the_input_box.toPlainText().startswith("System:"):
        time.sleep(0.3)

    while not the_main_window.state == "idle":
        time.sleep(0.3)

    response = the_input_box.toPlainText()

    return jsonify({"response": response})



@app.route("/tts", methods=["POST"])
def tts():
    """
    This function receives a text to speech request from the user and returns the response.
    """
    from .gpt_computer_assistant import the_main_window, the_input_box
    original_tts = the_main_window.tts_available
    the_main_window.tts_available = True
    the_main_window.manuel_stop = True
    data = request.json
    text = data["text"]
    print("TTS:", text)
    from .agent.process import tts_if_you_can
    tts_if_you_can(text, not_threaded=True, status_edit=True)
    the_main_window.tts_available = original_tts

    return jsonify({"response": "TTS request received"})

@app.route("/profile", methods=["POST"])
def profile():
    """
    This function sets the profile for the application.
    """
    data = request.json
    profile = data["profile"]
    print("Profile:", profile)
    from .utils.db import set_profile
    set_profile(profile)
    from .gpt_computer_assistant import the_main_window
    the_main_window.update_from_thread("Profile set to "+profile)
    return jsonify({"response": "Profile set to "+profile})


@app.route("/reset_memory", methods=["POST"])
def reset_memory():
    """
    This function resets the memory of the application.
    """
    from .agent.chat_history import clear_chat_history
    clear_chat_history()
    from .gpt_computer_assistant import the_main_window
    the_main_window.update_from_thread("Memory reset")
    return jsonify({"response": "Memory reset"})



@app.route("/activate_predefined_agents", methods=["POST"])
def enable_predefined_agents():
    """
    This function enables predefined agents for the application.
    """
    from .utils.db import activate_predefined_agents_setting
    activate_predefined_agents_setting()
    from .gpt_computer_assistant import the_main_window
    the_main_window.update_from_thread("Predefined agents enabled")
    return jsonify({"response": "Predefined agents enabled"})

@app.route("/deactivate_predefined_agents", methods=["POST"])
def disable_predefined_agents():
    """
    This function disables predefined agents for the application.
    """
    from .utils.db import deactivate_predefined_agents_setting
    deactivate_predefined_agents_setting()
    from .gpt_computer_assistant import the_main_window
    the_main_window.update_from_thread("Predefined agents disabled")
    return jsonify({"response": "Predefined agents disabled"})


@app.route("/activate_online_tools", methods=["POST"])
def enable_online_tools():
    """
    This function enables online tools for the application.
    """
    from .utils.db import activate_online_tools_setting
    activate_online_tools_setting()
    from .gpt_computer_assistant import the_main_window
    the_main_window.update_from_thread("Online tools enabled")
    return jsonify({"response": "Online tools enabled"})



@app.route("/deactivate_online_tools", methods=["POST"])
def disable_online_tools():
    """
    This function disables online tools for the application.
    """
    from .utils.db import deactivate_online_tools_setting
    deactivate_online_tools_setting()
    from .gpt_computer_assistant import the_main_window
    the_main_window.update_from_thread("Online tools disabled")
    return jsonify({"response": "Online tools disabled"})




@app.route("/change_name", methods=["POST"])
def change_name():
    """
    This function changes the name of the application.
    """
    data = request.json
    new_name = data["new_name"]
    print("Name:", new_name)
    from .character import change_name
    change_name(new_name)
    return jsonify({"response": "Name changed to "+new_name})


@app.route("/change_developer", methods=["POST"])
def change_developer():
    """
    This function changes the developer of the application.
    """
    data = request.json
    new_developer = data["new_developer"]
    print("Developer:", new_developer)
    from .character import change_developer
    change_developer(new_developer)
    return jsonify({"response": "Developer changed to "+new_developer})




@app.route("/library_install", methods=["POST"])
def library_install():
    """
    This function install a library.
    """
    data = request.json
    library = data["library"]
    print("Library İnstall:", library)
    from .utils.pypi import install_library
    if install_library(library):
        return jsonify({"response": f"Library {library} installed"})
    else:
        return jsonify({"response": f"Library {library} installation failed"})

@app.route("/library_uninstall", methods=["POST"])
def library_uninstall():
    """
    This function uninstall a library.
    """
    data = request.json
    library = data["library"]
    print("Library Uninstall:", library)
    from .utils.pypi import uninstall_library
    if uninstall_library(library):
        return jsonify({"response": f"Library {library} uninstalled"})
    else:
        return jsonify({"response": f"Library {library} uninstallation failed"})


@app.route("/custom_tool", methods=["POST"])
def custom_tool():
    """
    This function adds a custom tool to the application.
    """
    data = request.json
    code = data["code"]
    print("Custom Tool:", code)
    from .utils.function import string_to_function
    try:
        func = string_to_function(code)
        from .tooler import Tool
        Tool(func)
        return jsonify({"response": f"Custom tool {func.__name__} added"})
    except Exception as e:
        return jsonify({"response": f"Custom tool addition failed: {e}"}), 500





@app.route("/top_bar_activate", methods=["POST"])
def top_bar_activate():
    """
    This function serve an animation of top bar to show an operations especialy
    """
    from .gpt_computer_assistant import the_main_window
    data = request.json
    text = data["text"]
    
    the_main_window.active_border_animation(text)
    return jsonify({"response": "Activated top bar animation"})

@app.route("/top_bar_deactivate", methods=["POST"])
def top_bar_deactivate():
    """
    This function stop the top bar animation
    """
    from .gpt_computer_assistant import the_main_window
    data = request.json
    text = data["text"]
    the_main_window.deactive_border_animation(text)
    return jsonify({"response": "Deactivated top bar animation"})



@app.route("/boop_sound", methods=["POST"])
def boop_sound():
    """
    This function sound an boop to user
    """

    from .gpt_computer_assistant import click_sound
    click_sound()
    return jsonify({"response": "Sound played"})


@app.route("/ask_to_user", methods=["POST"])
def ask_to_user():
    """
    This api asks question to the user and return the result
    """
    data = request.json
    question = data["question"]
    wait_for_answer = data["wait_for_answer"]
    from .standard_tools import ask_to_user
    result = ask_to_user(question, wait_for_answer)
    return jsonify({"response": result})


@app.route("/set_text", methods=["POST"])
def set_text():
    """
    This api set text to main window text input 
    """
    data = request.json
    text = data["text"]
    from .gpt_computer_assistant import the_main_window
    the_main_window.set_text_from_api(text)
    return jsonify({"response": "Text set."})


@app.route("/set_background_color", methods=["POST"])
def set_background_color():
    """
    This api set text to main window text input 
    """
    data = request.json
    color = data["color"]
    from .gpt_computer_assistant import the_main_window
    the_main_window.set_background_color(color)
    return jsonify({"response": "Background color set."})

@app.route("/set_opacity", methods=["POST"])
def set_opacity():
    """
    This api set text to main window text input 
    """
    data = request.json
    opacity = data["opacity"]
    from .gpt_computer_assistant import the_main_window
    the_main_window.set_opacity(opacity)
    return jsonify({"response": "Opacity set."})

@app.route("/set_border_radius", methods=["POST"])
def set_border_radius():
    """
    This api set text to main window text input 
    """
    data = request.json
    radius = data["radius"]
    from .gpt_computer_assistant import the_main_window
    the_main_window.set_border_radius(radius)
    return jsonify({"response": "Border radius set."})

@app.route("/collapse", methods=["POST"])
def collapse():
    """
    This api set text to main window text input 
    """
    from .gpt_computer_assistant import the_main_window
    the_main_window.collapse_gca_api()
    return jsonify({"response": "Collapsed."})

@app.route("/expand", methods=["POST"])
def expand():
    """
    This api set text to main window text input 
    """
    from .gpt_computer_assistant import the_main_window
    the_main_window.uncollapse_gca_api()
    return jsonify({"response": "Expanded."})


@app.route("/save_openai_api_key", methods=["POST"])
def save_openai_api_key():
    """
    This api saves the OpenAI API key
    """
    data = request.json
    openai_api_key = data["openai_api_key"]
    from .utils.db import save_api_key
    save_api_key(openai_api_key)
    return jsonify({"response": "OpenAI API key saved."})


@app.route("/save_openai_url", methods=["POST"])
def save_openai_url():
    """
    This api saves the OpenAI base URL
    """
    data = request.json
    openai_url = data["openai_url"]
    from .utils.db import save_openai_url
    save_openai_url(openai_url)
    return jsonify({"response": "OpenAI base URL saved."})

@app.route("/save_model_settings", methods=["POST"])
def save_model_settings():
    """
    This api saves the model settings
    """
    data = request.json
    model_settings = data["model_settings"]
    from .utils.db import save_model_settings
    save_model_settings(model_settings)
    return jsonify({"response": "Model settings saved."})

@app.route("/save_groq_api_key", methods=["POST"])
def save_groq_api_key():
    """
    This api saves the Groq API key
    """
    data = request.json
    groq_api_key = data["groq_api_key"]
    from .utils.db import save_groq_api_key
    save_groq_api_key(groq_api_key)
    return jsonify({"response": "Groq API key saved."})

@app.route("/save_google_api_key", methods=["POST"])
def save_google_api_key():
    """
    This api saves the Google Generative AI API key
    """
    data = request.json
    google_api_key = data["google_api_key"]
    from .utils.db import save_google_api_key
    save_google_api_key(google_api_key)
    return jsonify({"response": "Google Generative AI API key saved."})


@app.route("/save_tts_model_settings", methods=["POST"])
def save_tts_model_settings():
    """
    This api saves the TTS model settings
    """
    data = request.json
    tts_model_settings = data["tts_model_settings"]
    from .utils.db import save_tts_model_settings
    save_tts_model_settings(tts_model_settings)
    return jsonify({"response": "TTS model settings saved."})

@app.route("/save_stt_model_settings", methods=["POST"])
def save_stt_model_settings():
    """
    This api saves the STT model settings
    """
    data = request.json
    stt_model_settings = data["stt_model_settings"]
    from .utils.db import save_stt_model_settings
    save_stt_model_settings(stt_model_settings)
    return jsonify({"response": "STT model settings saved."})


@app.route("/show_logo", methods=["POST"])
def show_logo():
    """
    This api shows the custom logo
    """
    from .utils.db import activate_logo_active_setting
    activate_logo_active_setting()
    from .gpt_computer_assistant import the_main_window
    the_main_window.show_logo_api()
    return jsonify({"response": "Custom logo activated."})

@app.route("/hide_logo", methods=["POST"])
def hide_logo():
    """
    This api hides the custom logo
    """
    from .utils.db import deactivate_logo_active_setting
    deactivate_logo_active_setting()
    from .gpt_computer_assistant import the_main_window
    the_main_window.hide_logo_api()
    return jsonify({"response": "Custom logo deactivated."})


@app.route("/default_logo", methods=["POST"])
def default_logo():
    """
    This api enable default logo
    """
    from .utils.db import save_logo_file_path, icon_256_path, is_logo_active_setting_active

    save_logo_file_path(icon_256_path)

    from .gpt_computer_assistant import the_main_window
    the_main_window.tray_and_task_bar_logo_api()
    if is_logo_active_setting_active():
        the_main_window.show_logo_api()
    return jsonify({"response": "Custom logo deactivated."})

@app.route("/custom_logo_upload", methods=["POST"])
def custom_logo_upload():
    """
    This api uploads a custom logo
    """
    file = request.files["logo"]
    from .utils.db import save_logo_file_path, custom_logo_path, is_logo_active_setting_active
    file.save(custom_logo_path)
    save_logo_file_path(custom_logo_path)
    from .gpt_computer_assistant import the_main_window
    the_main_window.tray_and_task_bar_logo_api()
    if is_logo_active_setting_active():
        the_main_window.show_logo_api()
    return jsonify({"response": "Custom logo uploaded."})


@app.route("/activate_long_gca", methods=["POST"])
def activate_long_gca():
    """
    This api activates long GCA
    """
    from .gpt_computer_assistant import the_main_window
    the_main_window.activate_long_gca_api()
    return jsonify({"response": "Long GCA activated."})

@app.route("/deactivate_long_gca", methods=["POST"])
def deactivate_long_gca():
    """
    This api deactivates long GCA
    """
    from .gpt_computer_assistant import the_main_window
    the_main_window.deactivate_long_gca_api()
    return jsonify({"response": "Long GCA deactivated."})



@app.route("/train", methods=["POST"])
def train():
    """
    This api trains the gca with given url
    """
    data = request.json
    url = data["url"]
    from .utils.train import train
    the_result = train(url)
    return jsonify({"response": the_result})


class ServerThread(threading.Thread):
    def __init__(self, app, host, port):
        threading.Thread.__init__(self)
        self.srv = make_server(host, port, app)
        self.ctx = app.app_context()
        self.ctx.push()

    def run(self):
        print("Starting server")
        self.srv.serve_forever()

    def shutdown(self):
        print("Stopping server")
        self.srv.shutdown()

server_thread = None

def start_api():
    global server_thread
    if server_thread is None:
        server_thread = ServerThread(app, "localhost", 7541)
        server_thread.start()
        print("API started")
    else:
        print("API is already running")

def stop_api():
    global server_thread
    if server_thread is not None:
        server_thread.shutdown()
        server_thread.join()
        server_thread = None
        print("API stopped")
    else:
        print("API is not running")







try:
    from .agent.chat_history import *
    from .agent.assistant import *
    from .llm import *
    from .llm_settings import llm_settings
    from .agent.agent import *
    from .agent.background import *

    from .gui.signal import *
    from .gui.button import *
    from .gui.settings import settings_popup
    from .gui.llmsettings import llmsettings_popup
    from .utils.db import *
    from .utils.telemetry import my_tracer, os_name

    from .audio.wake_word import wake_word
    from .audio.tts import text_to_speech
    from .character import name, developer

except ImportError:
    # This is for running the script directly
    # in order to test the GUI without rebuilding the package
    from agent.chat_history import *
    from agent.assistant import *
    from llm import *
    from llm_settings import llm_settings
    from agent.agent import *
    from agent.background import *
    from utils.db import *
    from gui.signal import *
    from gui.button import *
    from gui.settings import settings_popup
    from gui.llmsettings import llmsettings_popup
    from utils.telemetry import my_tracer, os_name
    from audio.wake_word import wake_word
    from audio.tts import text_to_speech
import platform
import threading
import time
import random
import math

from PyQt5.QtWidgets import QApplication, QMainWindow, QVBoxLayout, QWidget
from PyQt5.QtGui import QMouseEvent, QPainter, QPen, QBrush, QIcon, QColor
from PyQt5.QtCore import Qt, QTimer, QRect, pyqtSignal
from PyQt5.QtGui import QKeySequence
from PyQt5.QtWidgets import QShortcut
from PyQt5.QtWidgets import QSpacerItem, QSizePolicy
from PyQt5.QtWidgets import QDesktopWidget


from PyQt5.QtWidgets import (
    QPushButton,
    QLabel,
    QHBoxLayout,
)
from PyQt5.QtCore import QPoint

from PyQt5.QtWidgets import QTextEdit
from PyQt5 import QtGui
from PyQt5.QtCore import QThread
import pygame


print("Imported all libraries")


from PyQt5 import QtCore


try:
    import ctypes

    myappid = "onuratakan.gpt_computer_assistant.gui.1"
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
except:
    pass

the_input_box = None
the_input_text = None


the_input_box_pre = None


the_main_window = None


user_id = load_user_id()
os_name_ = os_name()



from PyQt5.QtCore import Qt, QRegExp
from PyQt5.QtGui import QColor, QSyntaxHighlighter, QTextCharFormat, QFont
from PyQt5.QtWidgets import QApplication, QTextEdit

class PythonSyntaxHighlighter(QSyntaxHighlighter):
    def __init__(self, parent):
        super().__init__(parent)
        self.highlighting_rules = []

        # Define different text formats with appropriate colors
        keyword_format = QTextCharFormat()
        keyword_format.setForeground(QColor(127, 0, 85))  # Dark purple for keywords

        built_in_formats = QTextCharFormat()
        built_in_formats.setForeground(QColor(42, 0, 255))  # Dark blue for built-ins and constants

        string_format = QTextCharFormat()
        string_format.setForeground(QColor(0, 128, 0))  # Green for strings

        function_format = QTextCharFormat()
        function_format.setForeground(QColor(0, 0, 255))  # Blue for function names

        comment_format = QTextCharFormat()
        comment_format.setForeground(QColor(128, 128, 128))  # Gray for comments
        comment_format.setFontItalic(True)

        number_format = QTextCharFormat()
        number_format.setForeground(QColor(255, 0, 0))  # Red for numbers

        decorator_format = QTextCharFormat()
        decorator_format.setForeground(QColor(0, 0, 128))  # Navy blue for decorators

        # Markdown formatting
        header_format = QTextCharFormat()
        header_format.setForeground(QColor(0, 128, 128))  # Teal for headers
        header_format.setFontWeight(QFont.Bold)

        bold_format = QTextCharFormat()
        bold_format.setFontWeight(QFont.Bold)

        italic_format = QTextCharFormat()
        italic_format.setFontItalic(True)

        code_format = QTextCharFormat()
        code_format.setForeground(QColor(255, 140, 0))  # Dark orange for inline code
        code_format.setFontFamily("Courier New")
        code_format.setBackground(QColor(245, 245, 245))  # Light gray background for inline code

        block_code_format = QTextCharFormat()
        block_code_format.setForeground(QColor(255, 140, 0))  # Dark orange for code blocks
        block_code_format.setFontFamily("Courier New")
        block_code_format.setBackground(QColor(245, 245, 245))  # Light gray background for code blocks

        # Define the regular expressions
        keywords = [
            'def', 'class', 'if', 'else', 'elif', 'return', 'import', 'from', 'as', 'for', 
            'while', 'try', 'except', 'finally', 'with', 'async', 'await', 'yield', 'lambda', 
            'global', 'nonlocal', 'assert', 'del', 'pass', 'break', 'continue', 'and', 'or', 
            'not', 'is', 'in'
        ]
        self.highlighting_rules += [(QRegExp(r'\b' + word + r'\b'), keyword_format) for word in keywords]

        built_ins = [
            'True', 'False', 'None', '__init__', 'self', 'print', 'len', 'range', 'str', 'int', 
            'float', 'list', 'dict', 'set', 'tuple'
        ]
        self.highlighting_rules += [(QRegExp(r'\b' + word + r'\b'), built_in_formats) for word in built_ins]

        self.highlighting_rules.append((QRegExp(r'"[^"\\]*(\\.[^"\\]*)*"'), string_format))
        self.highlighting_rules.append((QRegExp(r"'[^'\\]*(\\.[^'\\]*)*'"), string_format))

        self.highlighting_rules.append((QRegExp(r'\bdef\b\s*(\w+)'), function_format))
        self.highlighting_rules.append((QRegExp(r'\bclass\b\s*(\w+)'), function_format))
        
        self.highlighting_rules.append((QRegExp(r'#.*'), comment_format))
        
        self.highlighting_rules.append((QRegExp(r'\b[0-9]+[lL]?\b'), number_format))
        self.highlighting_rules.append((QRegExp(r'\b0[xX][0-9A-Fa-f]+[lL]?\b'), number_format))
        self.highlighting_rules.append((QRegExp(r'\b0[oO]?[0-7]+[lL]?\b'), number_format))
        self.highlighting_rules.append((QRegExp(r'\b0[bB][01]+[lL]?\b'), number_format))

        self.highlighting_rules.append((QRegExp(r'@[^\s]+'), decorator_format))

        # Markdown rules
        self.highlighting_rules.append((QRegExp(r'^#{1,6} .+'), header_format))  # Headers
        self.highlighting_rules.append((QRegExp(r'\*\*[^*]+\*\*'), bold_format))  # **bold**
        self.highlighting_rules.append((QRegExp(r'__[^_]+__'), bold_format))  # __bold__
        self.highlighting_rules.append((QRegExp(r'\*[^*]+\*'), italic_format))  # *italic*
        self.highlighting_rules.append((QRegExp(r'_[^_]+_'), italic_format))  # _italic_
        self.highlighting_rules.append((QRegExp(r'`[^`]+`'), code_format))  # Inline code

    def highlightBlock(self, text):
        # Handle code blocks separately
        if text.strip().startswith("```"):
            self.setFormat(0, len(text), self.highlighting_rules[-1][1])
            return
        
        for pattern, format in self.highlighting_rules:
            expression = QRegExp(pattern)
            index = expression.indexIn(text)
            while index >= 0:
                length = expression.matchedLength()
                self.setFormat(index, length, format)
                index = expression.indexIn(text, index + length)

readed_sentences = []

import re
def split_with_multiple_delimiters(text, delimiters):
    """
    Splits the text by any of the given delimiters while keeping the delimiters in the resulting parts.

    :param text: The input text to be split.
    :param delimiters: A string of delimiters to split the text on.
    :return: A list of parts including the delimiters.
    """
    # Create a regular expression pattern that matches any of the delimiters
    pattern = re.compile(f'(.*?[{re.escape(delimiters)}])')
    parts = pattern.findall(text)

    # Check if the last part is not complete and remove it if necessary
    if (
        parts and text
        and not any(text.endswith(d) for d in delimiters)
        and parts
        and not any(parts[-1].endswith(d) for d in delimiters)
    ):
        parts.pop()

    return parts



def click_sound():
    pygame.mixer.init()


    retro = pygame.mixer.Sound(click_sound_path)
    retro.set_volume(0.1)
    retro.play()



class Worker(QThread):
    text_to_set = pyqtSignal(str)


    def __init__(self):
        super().__init__()
        self.the_input_text = None
        self.make_animation = True
        self.commited_text = []

    def run(self):
        while True:
            self.msleep(500)  # Simulate a time-consuming task

            if self.the_input_text:
                last_text = self.commited_text[-1] if len(self.commited_text) > 0 else ""
                if self.the_input_text != last_text:
                    self.commited_text.append(self.the_input_text)

                    if len(self.the_input_text) > 90 or not self.make_animation:
                        self.text_to_set.emit(self.the_input_text)
                    else:
                        for i in range(len(self.the_input_text)):
                            self.text_to_set.emit(self.the_input_text[:i + 1])
                            self.msleep(10)





return_key_event = None
class CustomTextEdit(QTextEdit):
    def __init__(self, parent=None):
        super(CustomTextEdit, self).__init__(parent)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            global return_key_event
            return_key_event()
        super(CustomTextEdit, self).keyPressEvent(event)  # Process other key events normally




class Worker_2(QThread):
    text_to_set = pyqtSignal(str)
    text_to_set_title_bar = pyqtSignal(str)


    def __init__(self):
        super().__init__()
        self.the_input_text = None
        self.title_bar_text = None
        self.prev = None
        self.commited_text = []

    def run(self):
        while True:
            self.msleep(500)  # Simulate a time-consuming task

            if self.the_input_text and (self.prev is None or self.prev != self.the_input_text):
                self.prev = self.the_input_text
                self.text_to_set.emit("True")
                for i in range(len(self.title_bar_text)):
                    self.text_to_set_title_bar.emit(self.title_bar_text[:i + 1])
                    self.msleep(10)    

            if not self.the_input_text and self.prev != self.the_input_text:
                self.prev = self.the_input_text
                self.text_to_set.emit("False")

                the_text = "  "+ name()

                for i in range(len(the_text)):
                    self.text_to_set_title_bar.emit(the_text[:i + 1])
                    self.msleep(10)                



class Worker_3(QThread):
    text_to_set = pyqtSignal(str)


    def __init__(self):
        super().__init__()
        self.the_input_text = None



    def run(self):
        while True:
            self.msleep(500)  # Simulate a time-consuming task

            if self.the_input_text:
                self.text_to_set.emit("True")
                self.the_input_text = None




class Worker_collapse(QThread):
    text_to_set = pyqtSignal(str)


    def __init__(self):
        super().__init__()
        self.the_input_text = None



    def run(self):
        while True:
            self.msleep(500)  # Simulate a time-consuming task

            if self.the_input_text:
                self.text_to_set.emit("True")
                self.the_input_text = None
               
        

class Worker_uncollapse(QThread):
    text_to_set = pyqtSignal(str)


    def __init__(self):
        super().__init__()
        self.the_input_text = None



    def run(self):
        while True:
            self.msleep(500)  # Simulate a time-consuming task

            if self.the_input_text:
                self.text_to_set.emit("True")
                self.the_input_text = None
               




class Worker_show_logo(QThread):
    text_to_set = pyqtSignal(str)


    def __init__(self):
        super().__init__()
        self.the_input_text = None



    def run(self):
        while True:
            self.msleep(500)  # Simulate a time-consuming task

            if self.the_input_text:
                self.text_to_set.emit("True")
                self.the_input_text = None
               

class Worker_hide_logo(QThread):
    text_to_set = pyqtSignal(str)


    def __init__(self):
        super().__init__()
        self.the_input_text = None



    def run(self):
        while True:
            self.msleep(500)  # Simulate a time-consuming task

            if self.the_input_text:
                self.text_to_set.emit("True")
                self.the_input_text = None


class Worker_activate_long_gca(QThread):
    text_to_set = pyqtSignal(str)


    def __init__(self):
        super().__init__()
        self.the_input_text = None



    def run(self):
        while True:
            self.msleep(500)  # Simulate a time-consuming task

            if self.the_input_text:
                self.text_to_set.emit("True")
                self.the_input_text = None


class Worker_deactivate_long_gca(QThread):
    text_to_set = pyqtSignal(str)


    def __init__(self):
        super().__init__()
        self.the_input_text = None



    def run(self):
        while True:
            self.msleep(500)  # Simulate a time-consuming task

            if self.the_input_text:
                self.text_to_set.emit("True")
                self.the_input_text = None



class Worker_tray_and_task_bar_logo(QThread):
    text_to_set = pyqtSignal(str)


    def __init__(self):
        super().__init__()
        self.the_input_text = None



    def run(self):
        while True:
            self.msleep(500)  # Simulate a time-consuming task

            if self.the_input_text:
                self.text_to_set.emit("True")
                self.the_input_text = None



class DrawingWidget(QWidget):
    def __init__(self, parent=None):
        super(DrawingWidget, self).__init__(parent)
        # Set widget properties if needed, e.g., size

        self.main_ = parent
        self.active_button = ""

    def paintEvent(self, event):

        if llm_settings[load_model_settings()]["vision"] is True:
            self.main_.screen_available = True
        else:
            self.main_.screen_available = False



        self.main_.setAutoFillBackground(True)
        painter = QPainter(self)
        painter = painter

        painter.setRenderHint(QPainter.Antialiasing)
        painter.setPen(QPen(QColor("#000"), 1))
        painter.setBrush(QBrush(Qt.black, Qt.SolidPattern))

        center_x = 95
        center_y = 40

        if "talking" in self.main_.state:
            # Draw a pulsating circle with smooth easing animation
            radius_variation = 5 * (1 + math.sin(self.main_.pulse_frame * math.pi / 100))
            radius = 70 + radius_variation
            painter.drawEllipse(
                int(center_x - radius / 2),
                int(center_y - radius / 2),
                int(radius),
                int(radius),
            )
        elif self.main_.state == "thinking":
            # more slow pulsating circle with smooth easing animation
            radius_variation = 5 * (1 + math.sin(self.main_.pulse_frame * math.pi / 100))
            radius = 70 + radius_variation
            painter.drawEllipse(
                int(center_x - radius / 2),
                int(center_y - radius / 2),
                int(radius),
                int(radius),
            )

        else:
            radius = 70
            if self.main_.screen_available:
                painter.drawEllipse( # Main Button
                    int(center_x - radius / 2),
                    int(center_y - radius / 2),
                    int(radius),
                    int(radius),
                )


        
        self.main_.circle_rect = QRect(
            int(center_x - radius / 2),
            int(center_y - radius / 2),
            int(radius),
            int(radius),
        )



        if not self.main_.state == "thinking":
            painter.setPen(QPen(QColor("#01EE8A"), 1)) 
            if self.main_.screen_available:
                painter.drawEllipse( # Main BUtton Green Border
                    int(center_x - radius / 2),
                    int(center_y - radius / 2),
                    int(radius),
                    int(radius),
                )
        else:
            painter.setPen(QPen(QColor("#23538F"), 1))
            
            painter.drawEllipse(
                int(center_x - radius / 2),
                int(center_y - radius / 2),
                int(radius),
                int(radius),
            )



        painter.setPen(QPen(QColor("#000"), 1))

       

        small_center_x = 165
        small_center_y = 25
        small_radius = 30
        
        painter.drawEllipse( # Microphone bacground black
            int(small_center_x - small_radius / 2),
            int(small_center_y - small_radius / 2),
            int(small_radius),
            int(small_radius),
        )

        self.main_.small_circle_rect = QRect(
            int(small_center_x - small_radius / 2),
            int(small_center_y - small_radius / 2),
            int(small_radius),
            int(small_radius),
        )

        # Draw the icon inside the circle
        icon_size = small_radius * 2 // 3  # Adjust the icon size relative to the circle
        icon_rect = QRect(
            small_center_x - icon_size // 2,
            small_center_y - icon_size // 2,
            icon_size,
            icon_size,
        )
        self.main_.small_circle_recticon = QIcon(microphone_icon_path)
        self.main_.small_circle_recticon.paint(painter, icon_rect)


        small_center_x = 30
        small_center_y = 60
        small_radius = 30
        painter.drawEllipse(
            int(small_center_x - small_radius / 2),
            int(small_center_y - small_radius / 2),
            int(small_radius),
            int(small_radius),
        )

        self.main_.small_circle_left = QRect(
            int(small_center_x - small_radius / 2),
            int(small_center_y - small_radius / 2),
            int(small_radius),
            int(small_radius),
        )

        # Draw the icon inside the circle
        icon_size = small_radius * 2 // 3  # Adjust the icon size relative to the circle
        icon_rect = QRect(
            small_center_x - icon_size // 2,
            small_center_y - icon_size // 2,
            icon_size,
            icon_size,
        )
        self.main_.small_circle_lefticon = QIcon(audio_icon_path)
        self.main_.small_circle_lefticon.paint(painter, icon_rect)



        small_center_x = 30
        small_center_y = 25
        small_radius = 30
        if self.main_.screen_available:
            painter.drawEllipse( # ScreenShot BUtton
                int(small_center_x - small_radius / 2),
                int(small_center_y - small_radius / 2),
                int(small_radius),
                int(small_radius),
            )

        
        self.main_.small_circle_left_top = QRect(
            int(small_center_x - small_radius / 2),
            int(small_center_y - small_radius / 2),
            int(small_radius),
            int(small_radius),
        )



        self.main_.screenshot_button_coordinates_size = [
            int(small_center_x - small_radius / 2),
            int(small_center_y - small_radius / 2),
            int(small_radius),
            int(small_radius),
        ]

        if self.active_button == "screenshot":
            self.screenshot_button_border_activate(painter)
            self.active_button = ""


        if self.main_.screen_available:
            # Draw the icon inside the circle
            icon_size = small_radius * 2 // 3  # Adjust the icon size relative to the circle
            icon_rect = QRect(
                small_center_x - icon_size // 2,
                small_center_y - icon_size // 2,
                icon_size,
                icon_size,
            )
            self.main_.small_circle_left_topticon = QIcon(screenshot_icon_path)
            self.main_.small_circle_left_topticon.paint(painter, icon_rect)






        small_center_x = 165
        small_center_y = 60
        small_radius = 30
        painter.drawEllipse(
            int(small_center_x - small_radius / 2),
            int(small_center_y - small_radius / 2),
            int(small_radius),
            int(small_radius),
        )

        self.main_.small_circle_collapse = QRect(
            int(small_center_x - small_radius / 2),
            int(small_center_y - small_radius / 2),
            int(small_radius),
            int(small_radius),
        )

        # Draw the icon inside the circle
        icon_size = small_radius * 2 // 3  # Adjust the icon size relative to the circle
        icon_rect = QRect(
            small_center_x - icon_size // 2,
            small_center_y - icon_size // 2,
            icon_size,
            icon_size,
        )

        if is_collapse_setting_active():
            self.main_.small_circle_collapse_icon = QIcon(down_icon_path)



        if not is_collapse_setting_active() and is_long_gca_setting_active():
            self.main_.small_circle_collapse_icon = QIcon(up_icon_path)

        self.main_.small_circle_collapse_icon.paint(painter, icon_rect)


    def screenshot_button_border_activate(self, painter):
        # Add an white border to the circle
        painter.setPen(QPen(QColor("#FFF"), 1))
        # Draw the ellipse with the specified green border
        self.main_.screenshot_button_border = painter.drawEllipse(
            self.main_.screenshot_button_coordinates_size[0],
            self.main_.screenshot_button_coordinates_size[1],
            self.main_.screenshot_button_coordinates_size[2],
            self.main_.screenshot_button_coordinates_size[3],
        )
        painter.setPen(QPen(QColor("#000"), 1)) 




    def mousePressEvent(self, event: QMouseEvent):


        self.main_.old_position = event.globalPos()

        with my_tracer.start_span("mouse_press_event") as span:
            span.set_attribute("user_id", user_id)
            span.set_attribute("os_name", os_name_)
            if self.main_.state == "idle" or "talking" in self.main_.state:
                try:
                    if self.main_.circle_rect.contains(event.pos()):

                        if self.main_.state == "aitalking":
                            self.main_.stop_ai_talking()

                        else:
                            self.main_.screenshot_and_microphone_button_action()
                except:
                    traceback.print_exc()

                try:
                            if self.main_.small_circle_rect.contains(event.pos()):
                                if self.main_.state == "aitalking":
                                    self.main_.stop_ai_talking()

                                else: 
                                    click_sound()
                                    self.main_.button_handler.toggle_recording(no_screenshot=True)
                except:
                    traceback.print_exc()

                try:

                            if self.main_.small_circle_left.contains(event.pos()):
                                if self.main_.state == "aitalking":
                                    self.main_.stop_ai_talking()

                                else:       
                                    click_sound()                          
                                    self.main_.button_handler.toggle_recording(take_system_audio=True)
                except:
                    traceback.print_exc()

                try:

                            if self.main_.small_circle_left_top.contains(event.pos()):
                                if self.main_.state == "aitalking":
                                    self.main_.stop_ai_talking()

                                else:
                                    click_sound()
                                    self.active_button = "screenshot"
                                    self.update()
                                    self.main_.button_handler.just_screenshot()
                except:
                    traceback.print_exc()

            try:
                if self.main_.small_circle_collapse.contains(event.pos()):



                    if not is_collapse_setting_active():
                        if is_long_gca_setting_active():
                            self.main_.deactivate_long_gca()
                            self.main_.collapse_gca()
                        else:
                            self.main_.activate_long_gca()

                    else:
                        self.main_.uncollapse_gca()




                    self.main_.update()
            except:
                pass


from PyQt5.QtCore import QVariantAnimation

class MainWindow(QMainWindow):
    api_enabled = False
    tts_available = True


    def screenshot_and_microphone_button_action(self):
        click_sound()
        if llm_settings[load_model_settings()]["vision"] is True:

            self.button_handler.toggle_recording(dont_save_image=True)
        else:
            self.button_handler.toggle_recording(no_screenshot=True)

    def stop_ai_talking(self):
        self.manuel_stop = True
        self.stop_talking = True

    def __init__(self):
        super().__init__()

        self.background_color = "45, 45, 45"
        self.opacity = 250
        self.border_radius = 10


        print("API Enabled:", MainWindow.api_enabled)
        if MainWindow.api_enabled:
            try:
                from .api import start_api
                start_api()
            except:
                raise Exception("API could not be started, please install gpt-computer-assistant[api]")
        self.stop_talking = False
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)  # Remove the default title bar

        # Load the San Francisco font
        print("Loading font")
        print(font_dir)
        try:
            font_id = QtGui.QFontDatabase.addApplicationFont(font_dir)


            font_family = QtGui.QFontDatabase.applicationFontFamilies(font_id)[0]
            self.setFont(QtGui.QFont(font_family))
        except:
            print("Error loading font")


        self.state = "idle"
        self.pulse_timer = None

        self.button_handler = ButtonHandler(self)
        self.initUI()
        self.old_position = self.pos()





        self.collapse = is_collapse_setting_active()
        if self.collapse:
            self.collapse_window()

        global the_main_window
        the_main_window = self


        self.general_styling()

        if is_dark_mode_active():
            self.dark_mode()
        else:
            self.light_mode()


        self.wake_word_thread = None

        self.wake_word_active = False

        if load_pvporcupine_api_key() != "CHANGE_ME" and is_wake_word_active():
            self.wake_word_active = True
            self.wake_word_trigger()    


        self.manuel_stop = False    


        self.border_animation = None

        self.complated_answer = False


        self.reading_thread = False
        self.reading_thread_2 = False


        image_layout = QHBoxLayout()
        self.the_image = QLabel(self)
        self.the_image.setPixmap(QtGui.QPixmap(load_logo_file_path()).scaled(25, 25))

        image_layout.addWidget(self.the_image)
        self.layout.addLayout(image_layout)
        self.the_image.setAlignment(Qt.AlignCenter)
        self.the_image.setFixedHeight(35)


        # Logo Adding
        if not is_logo_active_setting_active():
            self.the_image.hide()


        
        self.update_screen()




    def put_location(self):
        if load_location_setting() == "right":
            self.put_window_to_right_side_of_screen()

    def init_border_animation(self):
        # Create a QVariantAnimation to handle color change
        border_animation = QVariantAnimation(
            self,
            valueChanged=self.update_border_color,
            startValue=QColor("#303030"),
            endValue=QColor("#23538F"),
            duration=2000  # Duration for one loop in milliseconds
        )
        border_animation.setLoopCount(-1)  # Loop indefinitely
        return border_animation

    def start_border_animation(self, status):
        print("FUNCTION TRİGGERED")
        if self.border_animation is None:
            self.border_animation = self.init_border_animation()

        status = status.lower() == "true"
        if status:
            self.border_animation.start()
        else:
            self.border_animation.stop()
            self.title_bar.setStyleSheet("background-color: #2E2E2E; color: white; border-style: solid; border-radius: 15px; border-width: 0px; color: #fff;")



    def update_border_color(self, color):
        self.title_bar.setStyleSheet(f"background-color: #2E2E2E; color: white; border-style: solid; border-radius: 15px; border-width: 2px; border-color: {color.name()}; color: #fff;")
        self.title_bar.setStyleSheet(f"background-color: #2E2E2E; color: white; border-style: solid; border-radius: 15px; border-width: 1px; border-color: {color.name()}; color: #fff;")

    # Existing methods...

    def general_styling(self, a=None):
        self.setAttribute(Qt.WA_TranslucentBackground)
        self.setStyleSheet(f"border-radius: {self.border_radius}px; background-color: rgba({self.background_color}, {self.opacity});")
        self.central_widget.setStyleSheet("border-style: solid; border-width: 1px; border-color: rgb(0,0,0,0);")

        self.input_box_style = "border-radius: 10px; border-bottom: 1px solid #01EE8A;"



        self.settingsButton_style = "border-radius: 5px; height: 25px; border-style: solid;"
        self.llmsettingsButton_style = "border-radius: 5px; height: 25px; border-style: solid;"

        self.btn_minimize.setStyleSheet("background-color: #2E2E2E; color: white; border-style: none;")
        self.btn_close.setStyleSheet("background-color: #2E2E2E; color: white; border-style: none;")


    def set_background_color(self, color):
        self.background_color = color
        self.worker_3.the_input_text = "True"

    def set_opacity(self, opacity):
        self.opacity = opacity
        self.worker_3.the_input_text = "True"

    def set_border_radius(self, radius):
        self.border_radius = radius
        self.worker_3.the_input_text = "True"


    def wake_word_trigger(self):
        self.wake_word_thread = threading.Thread(target=self.wake_word)
        self.wake_word_thread.start()       

    def wake_word(self):
        from .agent.process import tts_if_you_can
        while True and is_wake_word_active() and self.wake_word_active:
            if wake_word(self):

                def random_accept_words():
                    return random.choice(["Yes", "Sir", "Boss", "Master"])


                tts_if_you_can(random_accept_words(), not_threaded=True)


                def trigger_wake_word():
                    if is_wake_word_screen_setting_active() and llm_settings[load_model_settings()]["vision"]:
                        self.button_handler.toggle_recording(dont_save_image=True)
                    else:
                        self.button_handler.toggle_recording(no_screenshot=True)

                if self.state == "aitalking":
                    self.manuel_stop = True
                    self.stop_talking = True
                    time.sleep(1)
                    trigger_wake_word()
                    print("Stop talking")
                else:
                    trigger_wake_word()







    def dark_mode(self):
        self.setAutoFillBackground(True)
        p = self.palette()
        p.setColor(self.backgroundRole(), QColor("#171717"))  # Set background color to white
        self.setPalette(p)
        self.input_box.setStyleSheet(self.input_box_style+"background-color: #2E2E2E; color: white;")



        self.settingsButton.setStyleSheet(self.settingsButton_style+"background-color: #2E2E2E; color: white;")
        self.llmsettingsButton.setStyleSheet(self.llmsettingsButton_style+"background-color: #2E2E2E; color: white;")




    def light_mode(self):
        self.setAutoFillBackground(True)
        p = self.palette()
        p.setColor(self.backgroundRole(), QColor("#F0F0F0"))
        self.setPalette(p)
        self.input_box.setStyleSheet(self.input_box_style+"background-color: #FFFFFF; color: black;")

        self.settingsButton.setStyleSheet(self.settingsButton_style+"background-color: #FFFFFF; color: black; ")
        self.llmsettingsButton.setStyleSheet(self.llmsettingsButton_style+"background-color: #FFFFFF; color: black; ")





    def collapse_window(self):
        the_input_box.hide()

        self.settingsButton.hide()
        self.llmsettingsButton.hide()

        self.update_screen()



  




    def initUI(self):
        self.setWindowTitle("GPT")
        self.setGeometry(100, 100, 200, 200)
        width = 210
        height = 300
  
        # setting the minimum size 
        self.setMinimumSize(width, height) 

        self.first_height = self.height()
        self.first_width = self.width()



        self.central_widget = QWidget(self)
        self.setCentralWidget(self.central_widget)
        layout = QVBoxLayout(self.central_widget)

        # Custom title bar
        self.title_bar = QWidget(self)
        self.title_bar.setFixedHeight(30)  # Set a fixed height for the title bar
        self.title_bar.setStyleSheet("background-color: #2E2E2E; color: #fff; border-radius: 15px; border-style: solid; border-width: 1px; border-color: #303030;")

        self.title_bar_layout = QHBoxLayout(self.title_bar)
        self.title_bar_layout.setContentsMargins(5, 5, 0, 5)
        self.title_bar_layout.setSpacing(0)

        self.btn_minimize = QPushButton("_", self.title_bar)
        self.btn_minimize.setFixedSize(20, 20)
        self.btn_minimize.clicked.connect(self.showMinimized)

        def stop_app():
            self.stop_talking = True
            self.wake_word_active = False
            if MainWindow.api_enabled:
                from .api import stop_api
                stop_api()
            self.close()


        self.btn_close = QPushButton("X", self.title_bar)
        self.btn_close.setFixedSize(20, 20)
        self.btn_close.clicked.connect(stop_app)

        self.title_label = QLabel("  "+name(), self.title_bar)

        # Change font size
        font = QtGui.QFont()
        font.setPointSize(11)
        self.title_label.setFont(font)

        self.title_label.setStyleSheet("border: 0px solid blue;") 

        self.title_bar_layout.addWidget(self.title_label)
        self.title_bar_layout.addStretch()
        self.title_bar_layout.addWidget(self.btn_minimize)



        self.title_bar_layout.addWidget(self.btn_close)


        # Create a spacer item with expanding policy
        spacer = QSpacerItem(5, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.title_bar_layout.addSpacerItem(spacer)  # Add spacer to the layout



        layout.addWidget(self.title_bar)




        self.drawing_widget = DrawingWidget(self)
        layout.addWidget(self.drawing_widget)






        self.layout = layout

        self.setLayout(layout)



        # Add keyboard shortcuts
        self.shortcut_screenshot = QShortcut(QKeySequence("Ctrl+1"), self)
        self.shortcut_screenshot.activated.connect(
            lambda: self.button_handler.just_screenshot()
        )
        self.shortcut_screenshot = QShortcut(QKeySequence("Ctrl+2"), self)
        self.shortcut_screenshot.activated.connect(
            lambda: self.button_handler.toggle_recording(take_system_audio=True)
        )

        self.shortcut_no_screenshot = QShortcut(QKeySequence("Ctrl+e"), self)
        self.shortcut_no_screenshot.activated.connect(
            lambda: self.button_handler.toggle_recording(take_system_audio=True)
        )

        self.shortcut_no_screenshot = QShortcut(QKeySequence("Ctrl+3"), self)
        self.shortcut_no_screenshot.activated.connect(
            lambda: self.button_handler.toggle_recording(no_screenshot=True)
        )

        # I want to create an input box to bottom left and a send button to bottom right

        input_box = CustomTextEdit(self)
        self.input_box = input_box


        input_box.setFixedHeight(80)

        # Set text wrapping. I dont wat to cut the text
        input_box.setWordWrapMode(QtGui.QTextOption.NoWrap)

        # Change the font size 
        font = QtGui.QFont()
        font.setPointSize(12)
        input_box.setFont(font)

        self.highlighter = PythonSyntaxHighlighter(self.input_box.document())


        if load_api_key() == "CHANGE_ME":
            input_box.setPlaceholderText("Save your API Key, go to settings")
        else:

            if platform.system() == "Darwin":
                if llm_settings[load_model_settings()]["vision"] is False:
                    input_box.setPlaceholderText("Type here \nsand ↵ ")
                else:
                    input_box.setPlaceholderText("Type here \nand ↵ \nor ⌘ + ↵ (+screenshot)")
            else:
                if llm_settings[load_model_settings()]["vision"] is False:
                    input_box.setPlaceholderText("Type here \nand ↵ ")
                else:
                    input_box.setPlaceholderText("Type here \nand ↵ \nor Ctrl + ↵ (+screenshot)")
            # Add an information and use enter icon to the input box for mac            
        input_box.setGeometry(30, self.height() - 60, 200, 80)
        global the_input_box
        the_input_box = input_box

        def input_box_send():
            if input_box.toPlainText() != "":
                click_sound()
                self.button_handler.input_text(input_box.toPlainText())

        def input_box_send_screenshot():
            if input_box.toPlainText() != "":
                click_sound()
                self.button_handler.input_text_screenshot(input_box.toPlainText())

        self.layout.addWidget(input_box)









        self.shortcut_enter = QShortcut(QKeySequence("Ctrl+Return"), self)
        self.shortcut_enter.activated.connect(input_box_send_screenshot)

        global return_key_event
        return_key_event = input_box_send



        button_layout_ = QHBoxLayout()

        self.settingsButton = QPushButton("Chat Settings", self)
        self.settingsButton.clicked.connect(settings_popup)

        self.llmsettingsButton = QPushButton("LLM Settings", self)
        self.llmsettingsButton.clicked.connect(llmsettings_popup)

        button_layout_.addWidget(self.settingsButton)
        button_layout_.addWidget(self.llmsettingsButton)
        self.layout.addLayout(button_layout_)



        self.worker = Worker()
        self.worker.text_to_set.connect(self.set_text)
        self.worker.start()

        self.worker_2 = Worker_2()
        self.worker_2.text_to_set.connect(self.start_border_animation)
        self.worker_2.text_to_set_title_bar.connect(self.set_title_bar_text)
        self.worker_2.start()

        self.worker_3 = Worker_3()
        self.worker_3.text_to_set.connect(self.general_styling)
        self.worker_3.start()


        self.worker_collapse = Worker_collapse()
        self.worker_collapse.text_to_set.connect(self.collapse_gca)
        self.worker_collapse.start()


        self.worker_uncollapse = Worker_uncollapse()
        self.worker_uncollapse.text_to_set.connect(self.uncollapse_gca)
        self.worker_uncollapse.start()


        self.worker_show_logo = Worker_show_logo()
        self.worker_show_logo.text_to_set.connect(self.show_logo)
        self.worker_show_logo.start()

        self.worker_hide_logo = Worker_hide_logo()
        self.worker_hide_logo.text_to_set.connect(self.hide_logo)
        self.worker_hide_logo.start()



        self.worker_activate_long_gca = Worker_activate_long_gca()
        self.worker_activate_long_gca.text_to_set.connect(self.activate_long_gca)
        self.worker_activate_long_gca.start()

        self.worker_deactivate_long_gca = Worker_deactivate_long_gca()
        self.worker_deactivate_long_gca.text_to_set.connect(self.deactivate_long_gca)
        self.worker_deactivate_long_gca.start()


        self.worker_tray_and_task_bar_logo = Worker_tray_and_task_bar_logo()
        self.worker_tray_and_task_bar_logo.text_to_set.connect(self.tray_and_task_bar_logo)
        self.worker_tray_and_task_bar_logo.start()


        # print height and width
        print(self.height(), self.width())

        self.show()





    def set_text(self, text):
        global the_input_box


        vertical_scrollbar = the_input_box.verticalScrollBar()
        scroll_value = vertical_scrollbar.value()



        the_input_box.setPlainText(text)

        vertical_scrollbar.setValue(scroll_value)

    def set_title_bar_text(self, text):
        self.title_label.setText(text)

    def update_from_thread(self, text, system=True):
        self.worker.make_animation = True
        if system:
            text = "System: " + text
        print("Updating from thread", text)
        self.worker.the_input_text = text

    def read_part_task_generate_only(self):
        if not is_just_text_model_active() and the_main_window.tts_available:


            threads = {}

            the_okey_parts = split_with_multiple_delimiters(self.worker.the_input_text,".?!:")




            for each in the_okey_parts:
                if the_main_window.stop_talking:
                    break

                
                the_thread = threading.Thread(target=text_to_speech, args=(each,))

                threads[each] = the_thread
                the_thread.start()


            for each in threads.values():
                each.join()


        self.reading_thread_2 = False

    def read_part_task(self):
        if not is_just_text_model_active() and the_main_window.tts_available:
            threads = {}

            the_okey_parts = split_with_multiple_delimiters(self.worker.the_input_text,".?!:")


            will_read_parts = []

            for each in the_okey_parts:
                if the_main_window.stop_talking:
                    break
                if each not in readed_sentences:
                    will_read_parts.append(each)
                    readed_sentences.append(each)

                    the_thread = threading.Thread(target=text_to_speech, args=(each,))

                    threads[each] = the_thread
                    the_thread.start()



            for each in will_read_parts:
                    if the_main_window.stop_talking:
                        break
                    threads[each].join()

                    tts_if_you_can(each, not_threaded=True, bypass_other_settings=True)


        self.reading_thread = False


    def set_text_to_input_box(self, text):
        global readed_sentences
        self.worker.make_animation = False
        if self.worker.the_input_text.startswith("System:") or self.complated_answer:
            self.worker.the_input_text = ""
            self.complated_answer = False
            readed_sentences = []
        if text not in (">", "<>", ">\n", "<", "<\n"):

            self.worker.the_input_text += text

            if self.reading_thread is not True and len(self.worker.the_input_text) > 40:
                self.reading_thread = True
                threading.Thread(target=self.read_part_task).start()

            if self.reading_thread_2 is not True and len(self.worker.the_input_text) > 250:
                self.reading_thread_2 = True
                threading.Thread(target=self.read_part_task_generate_only).start()

        else:
            print("Problem on text chars")




    def set_text_from_api(self, text):
        self.worker.make_animation = True
        self.worker.the_input_text = text










    def active_border_animation(self, title_bar_text = None):
        if self.worker_2.title_bar_text is not None:
            if self.worker_2.title_bar_text != title_bar_text:
                return

        self.worker_2.the_input_text = True
        if title_bar_text is None:
            title_bar_text = "  "+name()
        else:
            title_bar_text = f"  {title_bar_text}"
            if len(title_bar_text) > 33:
                title_bar_text = title_bar_text[:30] + "..."
        self.worker_2.title_bar_text = title_bar_text

        self.btn_minimize.hide()
        self.btn_close.hide()
    def deactive_border_animation(self, title_bar_text=None):
        
        if title_bar_text is None:
            title_bar_text = "  "+name()
        else:
            title_bar_text = f"  {title_bar_text}"
            if len(title_bar_text) > 33:
                title_bar_text = title_bar_text[:30] + "..."

        if self.worker_2.title_bar_text is not None:
            if self.worker_2.title_bar_text != title_bar_text:
                return

        self.worker_2.the_input_text = False
        self.worker_2.title_bar_text = None
        time.sleep(1)
        self.btn_minimize.show()
        self.btn_close.show()


    def mouseMoveEvent(self, event: QMouseEvent):
        delta = QPoint(event.globalPos() - self.old_position)
        if event.buttons() == Qt.LeftButton and self.title_bar.underMouse():
            self.move(self.x() + delta.x(), self.y() + delta.y())
            self.old_position = event.globalPos()


    def mousePressEvent(self, event: QMouseEvent):
        self.old_position = event.globalPos()


    def remove_screenshot_button(self):
        self.update()


    def add_screenshot_button(self):
        self.update()


    def update_state(self, new_state):

        assistant_stopped = False
        if self.state == "aitalking" and new_state == "idle":
            assistant_stopped = True

        if self.manuel_stop:
            assistant_stopped = False
            self.manuel_stop = False



        self.state = new_state
        print(f"State updated: {new_state}")
        if "talking" in new_state:
            self.tray.setIcon(self.tray_active_icon)
            self.pulse_frame = 0
            if self.pulse_timer:
                self.pulse_timer.stop()
                self.pulse_timer = None
            self.pulse_timer = QTimer(self)
            self.pulse_timer.timeout.connect(self.pulse_circle)
            self.pulse_timer.start(5)
        elif new_state == "thinking":

            the_main_window.update_from_thread("Thinking...")
            self.pulse_frame = 0
            if self.pulse_timer:
                self.pulse_timer.stop()
                self.pulse_timer = None
            self.pulse_timer = QTimer(self)
            self.pulse_timer.timeout.connect(self.pulse_circle)
            self.pulse_timer.start(20)
        elif self.pulse_timer:
            self.tray.setIcon(self.tray_icon)
            self.pulse_timer.stop()
            self.pulse_timer = None
        self.update()  # Trigger a repaint

        if assistant_stopped:
            
                global the_input_box
                if the_input_box.toPlainText().endswith("?") and is_continuously_conversations_setting_active():
                    
                    self.button_handler.toggle_recording(no_screenshot=True, new_record=True)

        if new_state == "idle":
            click_sound()

    def pulse_circle(self):
        self.pulse_frame = (self.pulse_frame + 1) % 100
        self.update()




    def collapse_gca(self):
        self.collapse = True
        self.collapse_window()
        activate_collapse_setting()
        
        self.update_screen()

    def collapse_gca_api(self):
        self.worker_collapse.the_input_text = "True"


    def uncollapse_gca(self):
        self.collapse = False
        print()
        # hide all buttons and input box
        the_input_box.show()

        self.settingsButton.show()
        self.llmsettingsButton.show()




        deactivate_collapse_setting()

        self.update_screen()


    def uncollapse_gca_api(self):
        self.worker_uncollapse.the_input_text = "True"

        



    def show_logo(self):
        self.the_image.setPixmap(QtGui.QPixmap(load_logo_file_path()).scaled(25, 25))
        self.the_image.show()

        self.update_screen()



    def tray_and_task_bar_logo(self):
        app_icon = QtGui.QIcon()

        app_icon.addFile(load_logo_file_path(), QtCore.QSize(48, 48))




        self.the_app.setWindowIcon(app_icon)

        self.tray.setIcon(app_icon)

        self.tray_icon = app_icon
        self.tray_active_icon = app_icon

        print("ICON Set", load_logo_file_path())


    def tray_and_task_bar_logo_api(self):
        self.worker_tray_and_task_bar_logo.the_input_text = "True"


    def show_logo_api(self):
        self.worker_show_logo.the_input_text = "True"

    def hide_logo(self):
        self.the_image.hide()
        self.update_screen()
    
    def hide_logo_api(self):
        self.worker_hide_logo.the_input_text = "True"





    def activate_long_gca(self):
        activate_long_gca_setting()


        self.update_screen()

    def activate_long_gca_api(self):
        self.worker_activate_long_gca.the_input_text = "True"

    def deactivate_long_gca(self):
        deactivate_long_gca_setting()

        self.update_screen()

    def deactivate_long_gca_api(self):
        self.worker_deactivate_long_gca.the_input_text = "True"


    def update_screen(self):
        width = 210
        height = 320
  
        

        if is_logo_active_setting_active():
            height += 35

        if is_collapse_setting_active():

            height = 150
            if is_logo_active_setting_active():
                height += 35

        
        if is_long_gca_setting_active():
            if not is_collapse_setting_active():
                height += 500
                self.input_box.setFixedHeight(580)

        else:
            self.input_box.setFixedHeight(80)

            

        self.setFixedSize(width, height) 
        self.put_location()

        



    def put_window_to_right_side_of_screen(self):
        screen = QDesktopWidget().screenGeometry()
        window = self.frameGeometry()
        
        # Calculate x position for the right side of the screen and center vertically
        x = screen.width() - window.width()  # To right side
        y = (screen.height() - window.height()) // 2  # Center vertically


        # Add a small offset to the right side
        x -= 10

        self.move(x, y)




import struct

try:

    from ..utils.db import load_pvporcupine_api_key
except ImportError:
    from utils.db import load_pvporcupine_api_key

def wake_word(the_main_window):
    import pvporcupine
    import pyaudio

    porcupine = pvporcupine.create(access_key=load_pvporcupine_api_key(),
                                           keywords=pvporcupine.KEYWORDS)
    # Initialize PyAudio
    pa = pyaudio.PyAudio()

    # Open an audio stream
    audio_stream = pa.open(
        rate=porcupine.sample_rate,
        channels=1,
        format=pyaudio.paInt16,
        input=True,
        frames_per_buffer=porcupine.frame_length
    )

    print("Listening for wake word...")

    # Continuously listen for the wake word
    while the_main_window.wake_word_active:
        pcm = audio_stream.read(porcupine.frame_length)
        pcm = struct.unpack_from("h" * porcupine.frame_length, pcm)

        # Process the audio frame and check for the wake word
        keyword_index = porcupine.process(pcm)

        if keyword_index >= 0:
            print("Wake word detected!")
            return True



from PyQt5.QtCore import pyqtSignal, QObject



class SignalHandler(QObject):
    """
    A QObject subclass to handle signals used in the GUI application.

    This class defines several signals that can be used to communicate
    between different components of the GUI application.

    Signals:
    - recording_started: Signal emitted when recording is started.
    - recording_stopped: Signal emitted when recording is stopped.
    - assistant_thinking: Signal emitted when the assistant is processing a request.
    - assistant_response_ready: Signal emitted when the assistant response is ready to be displayed.
    - assistant_response_stopped: Signal emitted when the assistant response display is stopped.

    """

    recording_started = pyqtSignal()
    recording_stopped = pyqtSignal()
    assistant_thinking = pyqtSignal()
    assistant_response_ready = pyqtSignal()
    assistant_response_stopped = pyqtSignal()

signal_handler = SignalHandler()




import os
import platform
import sys
import webbrowser

from PyQt5.QtWidgets import QApplication
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import Qt
from pynput import keyboard


def start(api=False):
    """
    Starts the computer assistant application.

    This function starts the computer assistant application, which includes parsing command-line arguments
    to set the profile, initializing the graphical user interface, and starting the application event loop.

    Command-line Arguments:
    --profile (str): The profile to use for the application.

    Raises:
    ImportError: If the required modules or packages are not found.

    Returns:
    None
    """

    try:
        import crewai
    except:
        pass

    # get --profile argument with library
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--profile", help="profile to use")
    parser.add_argument("--api", help="Enable API mode", action="store_true")

    parser.add_argument("--set_tts_provider", help="Set tts provider only")
    parser.add_argument("--set_stt_provider", help="Set stt provider only")

    parser.add_argument("--set_llm", help="Set llm model only")


    args = parser.parse_args()

    set_tts_provider = args.set_tts_provider

    if set_tts_provider is not None:
        from .utils.db import save_tts_model_settings
        save_tts_model_settings(set_tts_provider)
        return
    
    set_stt_provider = args.set_stt_provider

    if set_stt_provider is not None:
        from .utils.db import save_stt_model_settings
        save_stt_model_settings(set_stt_provider)
        return
    
    set_llm = args.set_llm

    if set_llm is not None:
        from .utils.db import save_model_settings
        save_model_settings(set_llm)
        return


    profile = args.profile
    
    api_arg = args.api
    print("Profile:", profile)

    if profile is not None:
        from .utils.db import set_profile
        set_profile(profile)

    try:
        from .utils.db import load_tts_model_settings, load_stt_model_settings, is_logo_active_setting_active, load_logo_file_path
    except ImportError:
        from utils.db import load_tts_model_settings, load_stt_model_settings, is_logo_active_setting_active, load_logo_file_path

    if load_tts_model_settings() != "openai":
        from .audio.tts_providers.microsoft_local import preload_tts_microsoft_local
        preload_tts_microsoft_local()
    
    if load_stt_model_settings() != "openai":
        from .audio.stt_providers.openai_whisper_local import preload_stt_openai_whisper_local
        preload_stt_openai_whisper_local()
        


    try:
        from .gpt_computer_assistant import MainWindow
    except ImportError:
        from gpt_computer_assistant import MainWindow
    os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"

    if api or api_arg:
        print("API Enabled")
        MainWindow.api_enabled = True

    app = QApplication(sys.argv)
    ex = MainWindow()
    from PyQt5 import QtGui
    from PyQt5 import QtCore
    app_icon = QtGui.QIcon()

    app_icon.addFile(load_logo_file_path(), QtCore.QSize(48, 48))
    app.setWindowIcon(app_icon)

    ex.the_app = app

    # Create the tray
    menu_icon = QtGui.QIcon()
    menu_icon.addFile(load_logo_file_path(), QtCore.QSize(48, 48))

    menu_active_icon = QtGui.QIcon()
    menu_active_icon.addFile(load_logo_file_path(), QtCore.QSize(48, 48))

    tray = QSystemTrayIcon()
    tray.setIcon(menu_icon)
    tray.setVisible(True)
    ex.tray = tray
    ex.tray_active_icon = menu_active_icon
    ex.tray_icon = menu_icon

    # Create the menu
    menu = QMenu()


    ex.the_tray = tray





    show_menu = QAction("Show")
    def show_menu_connect():
        ex.setWindowState(Qt.WindowNoState)
    show_menu.triggered.connect(show_menu_connect)
    menu.addAction(show_menu)



    hide_menu = QAction("Hide")
    hide_menu.triggered.connect(ex.showMinimized)
    menu.addAction(hide_menu)

    menu.addSeparator()

    if platform.system() == "Darwin":
        the_text_of_screenshot_and_microphone = "Action: ⌃+⌥+⌘+up Screenshot and Microphone"
    else:
        the_text_of_screenshot_and_microphone = "Action: ctrl+alt+windows+up Screenshot and Microphone"
    screenshot_and_microphone = QAction(the_text_of_screenshot_and_microphone)
    def screenshot_and_microphone_connect():
        ex.setWindowState(Qt.WindowNoState)
        ex.screenshot_and_microphone_button_action()


    screenshot_listener = keyboard.GlobalHotKeys({
        '<ctrl>+<alt>+<cmd>+<up>': screenshot_and_microphone_connect
    })
    screenshot_listener.start()

    screenshot_and_microphone.triggered.connect(screenshot_and_microphone_connect)
    menu.addAction(screenshot_and_microphone)


    menu.addSeparator()

    action = QAction("Open GitHub Issues")
    action.triggered.connect(lambda: webbrowser.open("https://github.com/onuratakan/gpt-computer-assistant/issues"))
    menu.addAction(action)



    # Add a Quit option to the menu.
    quit = QAction("Quit")
    quit.triggered.connect(app.quit)
    menu.addAction(quit)

    # Add the menu to the tray
    tray.setContextMenu(menu)


    sys.exit(app.exec_())



import pyautogui
from .signal import *
import threading

try:
    from ..audio.record import *
    from ..screen.shot import *
    from ..agent.process import *
    from ..agent.chat_history import clear_chat_history
    from ..utils.db import (
        screenshot_path,
        save_api_key,
        load_api_key,
        activate_just_text_model,
        deactivate_just_text_model,
        is_just_text_model_active,
        set_profile,
        get_profile,
    )
    from ..screen.shot import take_screenshot
except ImportError:
    from audio.record import *
    from screen.shot import *
    from agent.process import *
    from agent.chat_history import clear_chat_history
    from utils.db import (
        screenshot_path,
        save_api_key,
        load_api_key,
        activate_just_text_model,
        deactivate_just_text_model,
        is_just_text_model_active,
        set_profile,
        get_profile,
    )
    from screen.shot import take_screenshot

recording_thread = None


class ButtonHandler:
    """Handles button click events and corresponding actions."""

    def __init__(self, main_window):
        """Initialize the ButtonHandler."""
        self.recording = False
        self.main_window = main_window
        self.process_audio_thread = None

        signal_handler.recording_started.connect(self.on_recording_started)
        signal_handler.recording_stopped.connect(self.on_recording_stopped)
        signal_handler.assistant_thinking.connect(self.on_assistant_thinking)
        signal_handler.assistant_response_ready.connect(
            self.on_assistant_response_ready
        )
        signal_handler.assistant_response_stopped.connect(
            self.on_assistant_response_stopped
        )

    def toggle_recording(
        self, no_screenshot=False, take_system_audio=False, dont_save_image=False, new_record=False
    ):
        """Toggle audio recording."""

        if self.recording and not new_record:
            stop_recording()
            self.recording = False
        else:
            if not no_screenshot:
                screenshot = pyautogui.screenshot()
                screenshot.save(screenshot_path)

            self.no_screenshot = no_screenshot
            self.take_system_audio = take_system_audio
            self.dont_save_image = dont_save_image

            global recording_thread
            if recording_thread is None or not recording_thread.is_alive() or new_record:
                recording_thread = threading.Thread(
                    target=start_recording, args=(take_system_audio,self,)
                )
                recording_thread.start()
            signal_handler.recording_started.emit()

    def on_recording_started(self):
        """Handle event when recording starts."""

        self.recording = True
        self.main_window.update_state("talking")

    def on_recording_stopped(self):
        """Handle event when recording stops."""

        print("ON RECORDING STOPPED")
        self.recording = False
        self.main_window.update_state("thinking")
        if (
            self.process_audio_thread is None
            or not self.process_audio_thread.is_alive()
        ):
            signal_handler.assistant_thinking.emit()
            self.process_audio_thread = threading.Thread(
                target=process_audio,
                args=(
                    not self.no_screenshot,
                    self.take_system_audio,
                    self.dont_save_image,
                ),
            )
            self.process_audio_thread.start()

    def just_screenshot(self):
        """Take a screenshot."""

        take_screenshot()
        self.process_audio_thread = threading.Thread(target=process_screenshot)
        self.process_audio_thread.start()

    def on_assistant_response_stopped(self):
        """Handle event when assistant's response stops."""

        self.main_window.update_state("idle")

    def on_assistant_thinking(self):
        """Handle event when assistant is thinking."""

        self.main_window.update_state("thinking")

    def on_assistant_response_ready(self):
        """Handle event when assistant's response is ready."""

        self.main_window.update_state("aitalking")

    def input_text(self, text):
        """Handle input text."""

        self.main_window.update_state("thinking")
        if (
            self.process_audio_thread is None
            or not self.process_audio_thread.is_alive()
        ):
            signal_handler.assistant_thinking.emit()
            self.process_audio_thread = threading.Thread(
                target=process_text, args=(text,)
            )
            self.process_audio_thread.start()

    def input_text_screenshot(self, text):
        """Handle input text with screenshot."""

        screenshot = pyautogui.screenshot()
        screenshot.save(screenshot_path)

        self.main_window.update_state("thinking")
        if (
            self.process_audio_thread is None
            or not self.process_audio_thread.is_alive()
        ):
            signal_handler.assistant_thinking.emit()
            self.process_audio_thread = threading.Thread(
                target=process_text,
                args=(text,),
                kwargs={"screenshot_path": screenshot_path},
            )
            self.process_audio_thread.start()
